// pltab.C Version 7.43  Copyleft (c) 杉本典夫
// Time-stamp: <2025-03-04 14:48:36 tac>
//
// パラメトリック多変量生命表解析
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W pltab pltab.C
//                -I/usr/localE/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "pltab  Version 7.43          Copyleft(Vo\\oV) 2025 杉本典夫",
  "書式:pltab [<オプション> …]  [<データファイル> …]",
  "機能:最初の項目を期間項目、次の項目を転帰項目、以後の項目を共変数とした",
  "     パラメトリック多変量生命表解析を行います。",
  "<オプション>",
  "  -b=<β0の初期値>[,<β1の初期値>,…]:偏回帰係数の初期値を指定します。",
  "  -cc[=<信頼係数>]:信頼係数を指定します。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -cin=<取り込み基準>:変数の取り込み基準Cinを指定します。(Cin≧0)",
  "  -d=<死亡コード>:死亡のコードを指定します。(オプション省略時:-d=0)",
  "  -dnsname[=<出力ファイル名>]:DANSデータを出力するファイル名を指定する。",
  "    オプション省略時:DANSデータ出力せず  -dnsname:-dnsname=pltab_dns.d",
  "  -eps[=<収束誤差>]:最尤法の収束誤差を指定する。(0<収束誤差)",
  "  -f=<ファイル名>:指定したファイルからオプションを読み込みます。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示します。(-h:-h=man)",
  "  -method[={del|dif}]:計算方法の指定です(-method:-method=dif)。",
  "  -model[={exp|wei}]:分布モデルの指定です。(-model:-model=exe)",
  "  -ngp={dea|hr|kmd|kms|sur}[,…]:Ngraph用データの指定。",
  "    dea:累積死亡率曲線  hr:ハザード比  kmd:累積死亡率曲線",
  "    kms:累積生存率曲線  sur:生存率関数  オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル名>]:Ngraph用データを出力するファイル名を指定する。",
  "    オプション省略時:Ngraph用データ出力せず  -ngpname:-ngpname=pltab_ngp.dat",
  "  -pin=<取り込み基準>:変数の取り込み基準Pinを指定します。(0<Pin≦1)",
  "  -res=[-][{def|all|allx|cindex|ltab|ps[<累積生存率>]|selx|stat|v}][,…]",
  "       :解析結果の出力項目を指定する。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    allx:全変数選択結果  cindex:C-index  ltab:Kaplan-Meier法による生命表",
  "    ps[<累積生存率>]:％生存時間  selx:変数選択結果  stat:基礎統計量",
  "    v:偏回帰係数初期値",
  "  -rmax=<最大回数>:反復計算の最大回数を指定します。(0≦最大回数)",
  "  -x<共変数番号>=<共変数の値>[,x<共変数番号>=<共変数の値>,…]:Ngraph用",
  "  -xrange=<最小値>-<最小値>:X軸の最小値と最大値を指定する。"
};

const string Title = "=== パラメトリック生存時間解析 ==="; // 表題
const string DNSNAMEdef = "pltab_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "pltab_ngp.dat"; // 省略時Ngraphファイル名
constexpr double CINdef = 2.0;  // 省略時χ2乗取り込み基準
constexpr double EPSdef = 1.0e-5; // 省略時収束誤差
constexpr double PINdef = 0.1;    // 省略時p値取り込み基準
constexpr double PSdef = 0.5;     // 省略時累積生存率
constexpr int RMAXdef = 200;      // 省略時反復計算最大回数

enum methodtype                 // 計算方法の種類
{
  methodnon,                    // 計算方法指定なし
  methoddel,                    // 差分法
  methoddif,                    // 微分法
  methodetc                     // その他
};

enum modeltype                  // 分布モデルの種類
{
  modelnon,                     // モデル指定なし
  modelexp,                     // 指数分布モデル
  modelwei,                     // Weibull分布モデル
  modeletc                      // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  lstr blst;                    // 偏回帰係数の初期値リスト
  vecdbl vbvec;                 // 偏回帰係数の初期値ベクトル
  double vcc;                   // 信頼係数
  double vcin;                  // χ2乗値の取り込み基準
  string vdcode;                // 死亡のコード
  string vdnsname;              // DANSファイル名
  double veps;                  // 最尤法の収束誤差
  methodtype vmethod;           // 計算方法の種類
  modeltype vmodel;             // 分布モデルの種類
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  mapsb vngpmap;                // Ngraph出力の有無マップ
  double vpin;                  // p値の取り込み基準
  double vps;                   // 生存時間を推定する累積生存率
  lstr vreslst;                 // 分析結果出力リスト
  mapsb vresmap;                // 分析結果出力の有無マップ
  int vrmax;                    // 反復計算の最大回数
  lstr xlst;                    // 共変数リスト
  vecdbl vxvec;                 // 共変数ベクトル
  pairdbl vxrange;              // X軸の最小値と最大値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double b(size_t idx) const    // 偏回帰係数初期値を返す
  { return (idx < vbvec.size()) ? vbvec[idx] : BLANKf; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double cin() const            // χ2乗値の取り込み基準を返す
  { return vcin; }
  const string& dcode() const   // 死亡のコードを返す
  { return vdcode; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  double eps() const            // 最尤法の収束誤差を返す
  { return veps; }
  methodtype method() const     // 計算方法の種類を返す
  { return vmethod; }
  modeltype model() const       // 分布モデルの種類を返す
  { return vmodel; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  double pin() const            // p値の取り込み基準を返す
  { return vpin; }
  double ps() const             // 生存時間を推定する累積生存率を返す
  { return vps; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 反復計算の最大回数を返す
  { return vrmax; }
  double x(size_t xidx) const     // 共変数の値を返す
  { return (xidx < vxvec.size()) ? vxvec[xidx] : BLANKf; }
  double xmin() const           // X軸の最小値を返す
  { return vxrange.first; }
  double xmax() const           // X軸の最大値を返す
  { return vxrange.second; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), blst(), vbvec(), vcc(CCdef), vcin(CINdef), vdcode("0"), vdnsname(), veps(EPSdef), vmethod(methoddif), vmodel(modelexp), vngpname(), vngplst(), vngpmap(), vpin(1.0), vps(PSdef), vreslst(), vresmap(), vrmax(RMAXdef), xlst(), vxvec(), vxrange(BLANKf, BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  mapsb_citr ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "b")            // 偏回帰係数初期値の指定
  {
    if (strsplit(blst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "cc")      // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "cin")     // χ2乗値の取り込み基準指定
  {
    vcin = opt[1].empty() ? CINdef : strtodf(opt[1]);
    if (vcin < 0.0) err = badopt(vpname, optstr, USAGE, 0.0, FDmax);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "d")       // 死亡のコード指定
  { vdcode = opt[1]; }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 最尤法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < FDeps) err = badopt(vpname, optstr, USAGE, FDeps, FDmax);
  }
  else if (opt[0] == "method")  // 計算方法の種類指定
  {
    if (opt[1].empty() || opt[1] == "dif") vmethod = methoddif;
    else if (opt[1] == "del") vmethod = methoddel;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "model")   // 分布モデルの種類指定
  {
    if (opt[1].empty() || opt[1] == "exp") vmodel = modelexp;
    else if (opt[1] == "wei") vmodel = modelwei;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pin")     // p値の取り込み基準指定
  {
    vpin = opt[1].empty() ? PINdef : strtodf(opt[1]);
    if (vpin <= 0.0 || 1.0 < vpin) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 反復計算の最大回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "xrange")  // X軸の最小値と最大値の指定
  {
    if (getrange(vxrange.first, vxrange.second, opt[1]) == 0) // 範囲指定無
    { err = badopt(vpname, optstr, USAGE, 0.0, FDmax); }
    else if (vxrange.first < 0.0 || vxrange.first >= vxrange.second) // 範囲指定不適切
    { err = badopt(vpname, optstr, USAGE, 0.0, FDmax); }
  }
  else if (opt[0][0] == 'x')    // 共変数の値指定
  {
    if (strsplit(xlst, optstr, CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 3) return disperr(vpname, "項目数が足りません！");
  if (set_ngp() < 0) return -2; // -ngpオプションのエラー
  if (set_res() < 0) return -3; // -resオプションのエラー
  int nx = ni - 1;
  int np = (vmodel == modelwei) ? nx + 1 : nx;
  if ((int)(blst.size()) > np) return disperr(vpname, "偏回帰係数初期値の値が多すぎます！");
  vbvec.resize(np);
  vbvec.assign(np, BLANKf);
  if (blst.size() > 0)
  {
    int i = 0;
    for (const string& bstr : blst)
    {
      double fdata = strtodf(bstr);
      if (fdata == BLANKf) return disperr(vpname, "初期値指定が間違っています→", bstr);
      vbvec[i] = fdata;
      ++i;
    }
  }
  int p = nx - 1;
  vxvec.resize(p);
  vxvec.assign(p, BLANKf);
  if (xlst.empty()) return 0;
  if ((int)(xlst.size()) > p) return disperr(vpname, "共変数の値が多すぎます！");
  for (const string& xstr : xlst)
  {
    string opt[2];
    if (getopt(opt, xstr, OPTdev) > 0)
    {
      char* ptr = 0;
      int xno = strtol(opt[0].substr(1).c_str(), &ptr, 0);
      if (xno < 1 || p < xno) return disperr(vpname, "共変数指定の項目番号が間違っています→", xstr);
      double fdata = strtodf(opt[1]);
      if (fdata == BLANKf) return disperr(vpname, "共変数指定の値が間違っています→", xstr);
      vxvec[xno - 1] = fdata;
    }
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["dea"] = vngpmap["hr"] = vngpmap["kmd"] = vngpmap["kms"] = vngpmap["sur"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["dea"] = vngpmap["hr"] = vngpmap["kmd"] = vngpmap["kms"] = vngpmap["sur"] = false;
    for (const string& ngpstr : vngplst)
    {
      mapsb_citr ngppos = vngpmap.find(ngpstr);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngpstr + "オプション指定が不適切！"); }
      else
      { vngpmap[ngpstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["selx"] = vresmap["stat"] = true;
  vresmap["allx"] = vresmap["cindex"] = vresmap["ltab"] = vresmap["ps"] = vresmap["v"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["selx"] = vresmap["stat"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr respos = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(resstr.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      { vresmap["def"] = vresmap["selx"] = vresmap["stat"] = true; }
      else if (resstr == "all") // 全て出力
      {
        vresmap["def"] = vresmap["allx"] = vresmap["cindex"] = vresmap["ltab"] = vresmap["ps"] = vresmap["selx"] = vresmap["stat"] = vresmap["v"] = true;
      }
      else if (resstr[0] == 'p' && resstr[1] == 's' && resstr.size() > 2)
      {
        vps = strtodf(resstr.substr(2)); // 累積生存率
        if (vps < 0.0 || 1.0 < vps) return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！");
        vresmap["ps"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vnx;                       // 共変数の数
  int vnp;                       // パラメータ数
  int vnc;                       // 例数
  int vnd;                       // 死亡の例数
  double vlamda0;                // 平均瞬間死亡率(平均ハザード)
  listlfm vlifemlst;             // 生命表リスト
  vector<statv<double>> statvec; // 共変数の基礎統計量ベクトル
  mat<double> vsmat;             // 積和行列
  vecdbl v_b0vec;                // 偏回帰係数初期値ベクトル
  vecdbl v_bvec;                 // 偏回帰係数ベクトル
  vecdbl v_vbvec;                // 偏回帰係数の分散ベクトル
  vecint vcheckvec;              // 変数選択の有無ベクトル
  int vnr;                       // 反復計算数
  double vll;                    // モデルの尤度
  double vl0;                    // 切片項の尤度
  double vlf;                    // 飽和モデルの尤度
  int vdfb;                      // 回帰モデルの自由度
  double vlreg;                  // 回帰モデルの尤度
  int vdfl;                      // ズレの自由度
  double vllof;                  // ズレの尤度
  int vdft;                      // 全体の自由度
  double vlt;                    // 全体の尤度
  restype vresvar;               // 計算結果の種類
  restype cal_csqr(option& opt, vecdbl& cvec, const vecint& checkvec, const vecdbl& bvec) const;
  // 変数選択用χ2乗値を計算する
  int sel_invar(option& opt, vecint& checkvec, vecdbl& btvec, vecdbl& vbtvec, const vecdbl& cvec, const vecdbl& bvec) const;
  // 取り込み変数の選択と各種設定
  int sel_outvar(option& opt, vecint& checkvec, vecdbl& bvec, vecdbl& vbvec, vecdbl& btvec, vecdbl& vbtvec) const;
  // 係数ベクトルの代入と追い出し変数の選択
  restype newton_pltab(option& opt, vecdbl& bvec, vecdbl& vbvec, int& nr, double& ll, const vecint& checkvec) const;
  // Newton法によるパラメトリックモデルの計算(最尤法)
  restype cal_lgh(option& opt, vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& bv) const;
  // 傾斜ベクトルgとHessian行列Hを計算する
  restype expdel_lgh(vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& hv) const;
  // 指数分布モデルの傾斜ベクトルgとHessian行列Hを差分法で計算する
  double exp_llfun(const vec<double>& hv, const vecint& checkvec) const;
  // 指数分布モデルの対数尤度を返す
  restype expdif_lgh(vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& bv) const;
  // 指数分布モデルの傾斜ベクトルgとHessian行列Hを微分法で計算する
  restype weidel_lgh(vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& hv) const;
  // Weibull分布モデルの傾斜ベクトルgとHessian行列Hを差分法で計算する
  double wei_llfun(const vec<double>& hv, const vecint& checkvec) const;
  // Weibull分布モデルの対数尤度を返す
  restype weidif_lgh(vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& bv) const;
  // Weibull分布モデルの傾斜ベクトルgとHessian行列Hを微分法で計算する
  void cal_pltab(option& opt); // 分散分析表用各種統計量を計算する
  int cal_pltablf(option& opt, double& lf, const vecint& checkvec) const;
  // 飽和モデルの尤度を求める
public:
  varclass(option& opt, size_t ni); // コンストラクタ
  ~varclass()                   // デストラクタ
  {}
  void cal(option& opt, ldata& datalst);
  // データを集計して生命表リスト等をセットする
  void stepsel(option& opt); // 増減法で変数を選択する
  void allsel(option& opt);  // 全変数を選択する
  int nx() const                // 共変数の数を返す
  { return vnx; }
  int np() const                // パラメータ数を返す
  { return vnp; }
  int nc() const                // 例数を返す
  { return vnc; }
  int nd() const                // 死亡例数を返す
  { return vnd; }
  int nr() const                // 反復計算数を返す
  { return vnr; }
  int nb() const                // 選択した変数の数を返す
  { return vcheckvec[vnx]; }
  int check(size_t idx) const   // 変数選択の有無を返す
  { return (idx < vcheckvec.size()) ? vcheckvec[idx] : 0; }
  double b0(size_t idx) const   // 偏回帰係数初期値を返す
  { return (idx < v_b0vec.size()) ? v_b0vec[idx] : 0.0; }
  double b(size_t idx) const    // 偏回帰係数を返す
  { return (idx < v_bvec.size()) ? v_bvec[idx] : 0.0; }
  double vb(size_t idx) const   // 偏回帰係数の分散を返す
  { return (idx < v_vbvec.size()) ? v_vbvec[idx] : 0.0; }
  double lamda0() const         // 平均瞬間死亡率を返す
  { return vlamda0; }
  double ll() const             // モデルの尤度を返す
  { return vll; }
  double l0() const             // 切片項の尤度を返す
  { return vl0; }
  double lf() const             // 飽和モデルの尤度を返す
  { return vlf; }
  int dfb() const               // 回帰モデルの自由度を返す
  { return vdfb; }
  double lreg() const           // 回帰モデルの尤度を返す
  { return vlreg; }
  int dfl() const               // ズレの自由度を返す
  { return vdfl; }
  double llof() const           // ズレの尤度を返す
  { return vllof; }
  int dft() const               // 全体の自由度を返す
  { return vdft; }
  double lt() const             // 全体の尤度を返す
  { return vlt; }
  const listlfm& lifemlst() const // 生命表リストを返す
  { return vlifemlst; }
  int n(size_t idx) const       // 共変数の例数を返す
  { return (idx < statvec.size()) ? (statvec[idx]).n() : 0; }
  double mean(size_t idx) const // 共変数の平均値を返す
  { return (idx < statvec.size()) ? (statvec[idx]).mean() : 0.0; }
  double vr(size_t idx) const   // 共変数の不偏分散を返す
  { return (idx < statvec.size()) ? (statvec[idx]).vr() : 0.0; }
  double sd(size_t idx) const   // 共変数の標準偏差を返す
  { return (idx < statvec.size()) ? (statvec[idx]).sd() : 0.0; }
  const mat<double>& smat() const // 積和行列を返す
  { return vsmat; }
  restype resvar() const        // 計算結果の種類を返す
  { return vresvar; }
};

varclass::varclass(option& opt, size_t ni) : vnx(ni - 1), vnp(0), vnc(0), vnd(0), vlamda0(0.0), vlifemlst(), statvec(ni), vsmat(ni), v_b0vec(), v_bvec(), v_vbvec(), vcheckvec(), vnr(0), vll(0.0), vl0(0.0), vlf(0.0), vdfb(0), vlreg(0.0), vdfl(0), vllof(0.0), vdft(0), vlt(0.0), vresvar(resnon)
// コンストラクタ
{
  vnp = (opt.model() == modelwei) ? ni : ni - 1;
  v_b0vec.resize(vnp);  v_bvec.resize(vnp);  v_vbvec.resize(vnp);
  v_b0vec.assign(vnp, 0.0);  v_bvec.assign(vnp, 0.0);  v_vbvec.resize(vnp, 0.0);
  vcheckvec.assign(vnp + 1, 0);
}

void varclass::cal(option& opt, ldata& datalst)
// データを集計して生命表リスト等をセットする
{
  vnd = 0;
  int dim = vnx + 1;
  for (int i = 0; i < dim; ++i) statvec[i].clear();
  vsmat = 0.0;
  vecdbl fdatavec(dim, 0.0);
  fdatavec[0] = 1.0;
  int cno = 1;                  // ケース番号
  for (auto& cdatai : datalst)
  {
    double t = cdatai.fdata(0);
    bool blank = (t == BLANKf) ? true : false;
    for (int j = 1; blank == false && j < vnx; ++j)
    {
      fdatavec[j] = cdatai.fdata(j + 1);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (blank == false)
    {
      vecdbl xvec(vnx - 1);
      statvec[0].sigma(t);
      for (int j = 1; j < vnx; ++j)
      { statvec[j].sigma(fdatavec[j]);  xvec[j - 1] = fdatavec[j]; }
      octype oc = ocalive;      // 生存
      if (cdatai.sdata(1) == opt.dcode()) { oc = ocdeath;  ++vnd; } // 死亡
      else if (cdatai.sdata(1) == BLANKs) oc = ocdropout; // 脱落
      lifem lifemi(cno, t, oc, xvec);
      vlifemlst.push_back(lifemi); // 生命表リストに追加する
      if (t > 0.0)              // 重回帰分析用集計
      {
        double lnt = log(t);
        statvec[dim - 1].sigma(lnt);  fdatavec[dim - 1] = lnt;
        for (int i = 0; i < dim; ++i)
        {
          double x = fdatavec[i];
          for (int j = i; j < dim; ++j) vsmat.add(i, j, x * fdatavec[j]);
        }
      }
    }
    ++cno;
  }
  vnc = vlifemlst.size();
  if (vnc < 1)                  // 生命表データなし
  {
    vresvar = reserr;
    return;
  }
  vlamda0 = (statvec[0].sum() > 0.0) ? (double)vnd / statvec[0].sum() : 0.0;
  vlifemlst.sort();             // 生命表リストをソートする
  int i = 0;
  double s = 1.0;
  for (auto& lifemi : vlifemlst)
  {
    if (lifemi.oc() == ocdeath) s *= (double)(vnc - i - 1) / (double)(vnc - i);
    lifemi.s(s);                // 累積生存率をセットする
    ++i;
  }
  for (int i = 0; i < dim - 1; ++i)
  {
    for (int j = i + 1; j < dim; ++j) vsmat.assign(j, i, vsmat(i, j));
  }
  vsmat.sweep(0);               // 積和行列に変換する
  vresvar = resok;              // 生命表作成成功
}

void varclass::stepsel(option& opt) // 増減法で変数を選択する
{
  vecint checkvec(vcheckvec);
  vecdbl bvec(v_bvec), vbvec(v_vbvec);
  vecdbl cvec(vnp, 0.0), btvec(bvec), vbtvec(vbvec);
  int iin = 0, iout = 0, nr = 0;
  vll = 0.0;  vnr = 0;
  checkvec[0] = 1;  checkvec[vnp] = 1; // aまたは切片は無条件に取り込み
  if (opt.model() == modelwei) { checkvec[1] = 1;  checkvec[vnp] = 2; }
  do {
    vresvar = cal_csqr(opt, cvec, checkvec, bvec);
    if (vresvar == resok)
    {
      iin = sel_invar(opt, checkvec, btvec, vbtvec, cvec, bvec);
      if (iin > 0)              // 取り込み変数あり
      {
        iout = 0;
        int i = 0;
        do {
          vresvar = newton_pltab(opt, btvec, vbtvec, nr, vll, checkvec);
          if (nr > vnr) vnr = nr; //最大反復回数を採用する
          if (vresvar != reserr)
          {
            i = sel_outvar(opt, checkvec, bvec, vbvec, btvec, vbtvec);
            if (i > 0) iout = i;    // 最後に追い出した変数番号を記憶
          }
        } while (vresvar == resok && i > 0); // 追い出し変数が無くなるまで続行
      }
    }
  } while (vresvar == resok && iin > 0 && iin != iout && checkvec[vnp] < vnp);
  v_bvec = bvec;  v_vbvec = vbvec;  vcheckvec = checkvec;
  cal_pltab(opt);               // 分散分析表用各種統計量を計算する
}

void varclass::allsel(option& opt) // 全変数を選択する
{
  for (int i = 1; i < vnx; ++i) vsmat.sweep(i); // 偏回帰係数計算
  int ks = 0, k = 0;
  if (opt.model() == modelwei)
  { ks = k = 1;  v_b0vec[0] = 0.1; } // ln(a)の初期値
  v_b0vec[ks] = log(vlamda0);  ++k;
  for (int i = 1; i < vnx; ++i, ++k)
  {
    v_b0vec[k] = -vsmat(i, vnx);
    if (fabs(v_b0vec[k]) < FDeps) v_b0vec[k] = 0.0;
    v_b0vec[ks] -= v_b0vec[k] * statvec[i].mean();
  }
  for (int i = 0; i < vnp; ++i) // オプションによる初期値
  {
    if (opt.b(i) != BLANKf) v_b0vec[i] = opt.b(i);
  }
  vecdbl bvec(v_b0vec), vbvec(v_vbvec);
  vecint checkvec(vnp + 1, 1);
  checkvec[vnp] = vnp;
  vll = 0.0;  vnr = 0;
  vresvar = newton_pltab(opt, bvec, vbvec, vnr, vll, checkvec);
  v_bvec = bvec;  v_vbvec = vbvec;  vcheckvec = checkvec;
  cal_pltab(opt);               // 分散分析表用各種統計量を計算する
}

restype varclass::cal_csqr(option& opt, vecdbl& cvec, const vecint& checkvec, const vecdbl& bvec) const
// 変数選択用χ2乗値を計算する
{
  int is = (opt.model() == modelwei) ? 2 : 1;
  vec<double> bv(bvec), gv(vnp);
  mat<double> hmat(vnp);
  for (int i = is; i < vnp; ++i)
  {
    if (checkvec[i] == 0) bv.assign(i, 0.0); // 未選択のパラメーターを0にする
  }
  double ll = 0.0;
  vecint tcheckvec(vnp + 1, 1);
  tcheckvec[vnp] = vnp;
  if (cal_lgh(opt, gv, hmat, ll, tcheckvec, bv) == reserr) return reserr; // 変数選択用χ2乗値計算できず
  hmat.inv();
  bv -= hmat * gv;
  for (int i = 0; i < vnp; ++i)
  { cvec[i] = (hmat(i, i) < 0.0) ? bv(i) * bv(i) / (-hmat(i, i)) : 0.0; }
  return resok;
}

int varclass::sel_invar(option& opt, vecint& checkvec, vecdbl& btvec, vecdbl& vbtvec, const vecdbl& cvec, const vecdbl& bvec) const
// 取り込み変数の選択と各種設定
{
  double cmax = 0.0;
  int imax = 0;
  int is = (opt.model() == modelwei) ? 2 : 1;
  for (int i = is; i < vnp; ++i)
  {
    if (checkvec[i] == 0 && cvec[i] > cmax) { cmax = cvec[i];  imax = i; }
  }
  if (imax < is) return 0;      // 取り込み変数なし
  checkvec[imax] = 1;  checkvec[vnp] += 1;
  btvec.resize(checkvec[vnp]);  vbtvec.resize(checkvec[vnp]);
  int j = 0;
  for (int i = 0; i < vnp; ++i)
  {
    if (checkvec[i])
    {
      btvec[j] = bvec[i];
      ++j;
    }
  }
  return imax;                  // 変数番号を返す
}

int varclass::sel_outvar(option& opt, vecint& checkvec, vecdbl& bvec, vecdbl& vbvec, vecdbl& btvec, vecdbl& vbtvec) const
// 係数ベクトルの代入と追い出し変数の選択
{
  double cmin = opt.cin();
  ldbl pmax = opt.pin();
  int imin = 0, j = 1, is = 1;
  bvec[0] = btvec[0];  vbvec[0] = vbtvec[0];
  if (opt.model() == modelwei)
  {
    j = 2;  is = 2;
    bvec[1] = btvec[1];  vbvec[1] = vbtvec[1];
  }
  bool csel = (opt.pin() < 1.0) ? false : true; // χ2乗値による変数選択
  for (int i = is; i < vnp; ++i)
  {
    if (checkvec[i])
    {
      bvec[i] = btvec[j];  vbvec[i] = vbtvec[j];
      double csqr = (vbtvec[j] > 0.0) ? btvec[j] * btvec[j] / vbtvec[j] : 0.0;
      if (csel)
      {
        if (csqr < cmin) { cmin = csqr;  imin = i; }
      }
      else
      {
        ldbl pr = ctop(csqr, 1);
        if (pr > pmax) { pmax = pr;  imin = i; }
      }
      ++j;
    }
  }
  if (imin < is) return 0;      // 追い出し変数なし
  checkvec[imin] = 0;  checkvec[vnp] -= 1;
  btvec.resize(checkvec[vnp]);  vbtvec.resize(checkvec[vnp]);
  j = 0;
  for (int i = 0; i < vnp; ++i)
  {
    if (checkvec[i])
    {
      btvec[j] = bvec[i];
      ++j;
    }
  }
  return imin;                  // 変数番号を返す
}

restype varclass::newton_pltab(option& opt, vecdbl& bvec, vecdbl& vbvec, int& nr, double& ll, const vecint& checkvec) const
// Newton法によるパラメトリックモデルの計算(最尤法)
{
  int npt = checkvec[vnp];      // 一時的な変数の数
  vec<double> bv(npt), gv(npt);
  mat<double> hmat(npt);
  for (int i = 0; i < npt; ++i) bv.assign(i, bvec[i]);
  double lold = 0.0;
  restype resnewton = resnon, reslgh = resnon;
  nr = 0;  ll = 0.0;
  while (resnewton == resnon && (reslgh = cal_lgh(opt, gv, hmat, ll, checkvec, bv)) == resok)
  {
    ++nr;  hmat.inv();
    if (fabs(ll - lold) < opt.eps()) resnewton = resok; // 収束
    else if (opt.rmax() > 0 && nr >= opt.rmax()) resnewton = resrmax; // 反復計算最大回数
    else { lold = ll;  bv -= hmat * gv; }
  }
  if (resnewton == resnon) return reslgh; // Newton法発散
  for (int i = 0; i < npt; ++i) { bvec[i] = bv(i);  vbvec[i] = -hmat(i, i); }
  return resnewton;             // Newton法成功
}

restype varclass::cal_lgh(option& opt, vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& bv) const
// 傾斜ベクトルgとHessian行列Hを計算する
{
  restype reslgh = resok;
  switch (opt.model())
  {
  case modelwei :
    if (opt.method() == methoddel) reslgh = weidel_lgh(gv, hmat, ll, checkvec, bv);
    else reslgh = weidif_lgh(gv, hmat, ll, checkvec, bv);
    break;
  default :
    if (opt.method() == methoddel) reslgh = expdel_lgh(gv, hmat, ll, checkvec, bv);
    else reslgh = expdif_lgh(gv, hmat, ll, checkvec, bv);
  }
  return reslgh;                // 計算の成否を返す
}

restype varclass::expdel_lgh(vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& hv) const
// 指数分布モデルの傾斜ベクトルgとHessian行列Hを差分法で計算する
{
  constexpr double Emax = 308.0; // 指数最大数
  int nh = hv.dim();            // パラメータ数
  for (int i = 0; i < nh; ++i)
  {
    if (isnan(hv(i)) || hv(i) >= Emax) return reserr; // Newton法発散
  }
  constexpr double Delta = 1.0e-5;  // 差分法の相対微小変動
  ll = exp_llfun(hv, checkvec);
  if (isnan(ll) || ll > 0.0 || ll < -FDmax) return reserr; // Newton法発散
  for (int i = 0; i < nh; ++i)
  {
    vec<double> hiv(hv);
    double dhi = (fabs(hv(i)) < FDeps) ? Delta : hv(i) * Delta;
    hiv.add(i, dhi);
    double lli = exp_llfun(hiv, checkvec);
    gv.assign(i, (lli - ll) / dhi);
    for (int j = i; j < nh; ++j)
    {
      vec<double> hjv(hv);
      double dhj = (fabs(hv(j)) < FDeps) ? Delta : hv(j) * Delta;
      hjv.add(j, dhj);
      double llj = exp_llfun(hjv, checkvec);
      vec<double> hijv(hiv);
      hijv.add(j, dhj);
      double llij = exp_llfun(hijv, checkvec);
      hmat.assign(i, j, (llij - lli - llj + ll) / dhi / dhj);
    }
  }
  for (int i = 0; i < nh - 1; ++i)
  {
    for (int j = i + 1; j < nh; ++j) hmat.assign(j, i, hmat(i, j));
  }
  return resok;                 // エラーなし
}

double varclass::exp_llfun(const vec<double>& hv, const vecint& checkvec) const
// 指数分布モデルの対数尤度を返す
{
  int npt = checkvec[vnp];
  vec<double> xv(npt);
  xv.assign(0, 1.0);
  double ll = 0.0;
  for (auto& lifemi : vlifemlst)
  {
    double ti = lifemi.t();
    int k = 0, j = 1;
    for (int i = 1; i < vnp; ++i, ++k)
    {
      if (checkvec[i])
      {
        xv.assign(j, lifemi.x(k));
        ++j;
      }
    }
    double lhi = inp(hv, xv);
   if (lifemi.oc() == ocdeath) ll += lhi; // 死亡例
   ll += -exp(lhi) * ti;
  }
  return ll;
}

restype varclass::expdif_lgh(vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& bv) const
// 指数分布モデルの傾斜ベクトルgとHessian行列Hを微分法で計算する
{
  constexpr double Emax = 308.0; // 指数最大値
  int npt = checkvec[vnp];
  double sdlh = 0.0, sth = 0.0;
  vec<double> sdxv(npt), sthxv(npt), xv(npt);
  restype reslgh = resok;
  xv.assign(0, 1.0);
  gv = 0.0;  hmat = 0.0;
  listlfm_citr litr = vlifemlst.begin();
  while (reslgh == resok && litr != vlifemlst.end())
  {
    int j = 1;
    for (int i = 1; i < vnp; ++i)
    {
      if (checkvec[i])
      {
        xv.assign(j, litr->x(i - 1));
        ++j;
      }
    }
    double lhi = inp(bv, xv);
    if (isnan(lhi) || fabs(lhi) >= Emax) // 発散
    { reslgh = reserr; }
    else
    {
      double hi = exp(lhi), ti = litr->t();
      sth += ti * hi;
      for (int j = 0; j < npt; ++j)
      {
        sthxv.add(j, ti * hi * xv(j));
        for (int k = j; k < npt; ++k) hmat.add(j, k, -ti * hi * xv(j) * xv(k));
      }
      if (litr->oc() == ocdeath)
      {
        sdlh += lhi;
        for (int j = 0; j < npt; ++j) sdxv.add(j, xv(j));
      }
      ++litr;
    }
  }
  if (reslgh == reserr) return reslgh; // 発散
  ll = sdlh - sth;  gv = sdxv - sthxv;
  for (int i = 0; i < npt - 1; ++i)
  {
    for (int j = i + 1; j < npt; ++j) hmat.assign(j, i, hmat(i, j));
  }
  return reslgh;                // 計算成功
}

restype varclass::weidel_lgh(vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& hv) const
// Weibull分布モデルの傾斜ベクトルgとHessian行列Hを差分法で計算する
{
  constexpr double Emax = 308.0; // 指数最大数
  int nh = hv.dim();            // パラメータ数
  for (int i = 0; i < nh; ++i)
  {
    if (isnan(hv(i)) || hv(i) >= Emax) return reserr; // Newton法発散
  }
  const double Delta = 1.0e-4;  // 差分法の相対微小変動
  ll = wei_llfun(hv, checkvec);
  if (isnan(ll) || ll > 0.0 || ll < -FDmax) return reserr; // Newton法発散
  for (int i = 0; i < nh; ++i)
  {
    vec hiv(hv);
    double dhi = (fabs(hv(i)) < FDeps) ? Delta : hv(i) * Delta;
    hiv.add(i, dhi);
    double lli = wei_llfun(hiv, checkvec);
    gv.assign(i, (lli - ll) / dhi);
    for (int j = i; j < nh; ++j)
    {
      vec<double> hjv(hv);
      double dhj = (fabs(hv(j)) < FDeps) ? Delta : hv(j) * Delta;
      hjv.add(j, dhj);
      double llj = wei_llfun(hjv, checkvec);
      vec<double> hijv(hiv);
      hijv.add(j, dhj);
      double llij = wei_llfun(hijv, checkvec);
      hmat.assign(i, j, (llij - lli - llj + ll) / dhi / dhj);
    }
  }
  for (int i = 0; i < nh - 1; ++i)
  {
    for (int j = i + 1; j < nh; ++j) hmat.assign(j, i, hmat(i, j));
  }
  return resok;                 // エラーなし
}

double varclass::wei_llfun(const vec<double>& hv, const vecint& checkvec) const
// Weibull分布モデルの対数尤度を返す
{
  int npt = checkvec[vnp];
  int nxt = npt - 1;
  vec<double> xv(nxt);
  xv.assign(0, 1.0);
  double ll = 0.0, a = exp(hv(0));
  for (auto& lifemi : vlifemlst)
  {
    double ti = lifemi.t();
    if (ti > 0.0)
    {
      int k = 0, j = 1;
      for (int i = 2; i < vnp; ++i, ++k)
      {
        if (checkvec[i])
        {
          xv.assign(j, lifemi.x(k));
          ++j;
        }
      }
      double lhi = 0.0;
      for (j = 0; j < nxt; ++j) lhi += hv(j + 1) * xv(j);
      if (lifemi.oc() == ocdeath) ll += a * lhi + hv(0) + (a - 1.0) * log(ti);
      ll += -pow(exp(lhi) * ti, a);
    }
  }
  return ll;
}

restype varclass::weidif_lgh(vec<double>& gv, mat<double>& hmat, double& ll, const vecint& checkvec, const vec<double>& bv) const
// Weibull分布モデルの傾斜ベクトルgとHessian行列Hを微分法で計算する
{
  constexpr double Emax = 308.0; // 指数最大値
  double la = bv(0);
  if (isnan(la) || fabs(la) >= Emax) return reserr; // 発散
  double a = exp(la);
  int npt = checkvec[vnp];
  int nxt = npt - 1;
  double sdlh = 0.0, sdlt = 0.0, shta = 0.0, shtalh = 0.0;
  double shtalh2 = 0.0, shtalhlt = 0.0, shtalt = 0.0, shtalt2 = 0.0;
  vec<double> sdxv(nxt), xv(nxt);
  int nd = 0;
  restype reslgh = resok;
  xv.assign(0, 1.0);
  gv = 0.0;  hmat = 0.0;
  listlfm_citr litr = vlifemlst.begin();
  while (reslgh == resok && litr != vlifemlst.end())
  {
    double ti = litr->t();
    if (ti > 0.0)
    {
      int k = 0, j = 1;
      for (int i = 2; i < vnp; ++i, ++k)
      {
        if (checkvec[i])
        {
          xv.assign(j, litr->x(k));
          ++j;
        }
      }
      double lhi = 0.0;
      for (j = 0; j < nxt; ++j) lhi += bv(j + 1) * xv(j);
      if (isnan(lhi) || fabs(lhi) >= Emax) // 発散
      { reslgh = reserr; }
      else
      {
        double hi = exp(lhi), lti = log(ti);
        if (litr->oc() == ocdeath) // 死亡例
        {
          ++nd;  sdlh += lhi;  sdlt += lti;
          for (j = 0; j < nxt; ++j) sdxv.add(j, xv(j));
        }
        double htai = pow(hi * ti, a);
        shta += htai;  shtalh += htai * lhi;
        shtalh2 += htai * lhi * lhi;  shtalhlt += htai * lhi * lti;
        shtalt += htai * lti;  shtalt2 += htai * lti * lti;
        for (j = 0; j < nxt; ++j)
        {
          gv.add(j + 1, -a * htai * xv(j));
          hmat.add(0, j + 1, -a * a * htai * lhi * xv(j) - a * htai * xv(j) - a * a * htai * lti * xv(j));
        }
        for (j = 0; j < nxt; ++j)
        {
          for (k = j; k < nxt; ++k) hmat.add(j + 1, k + 1, -a * a * htai * xv(j) * xv(k));
        }
      }
    }
    ++litr;
  }
  if (reslgh == reserr) return reslgh; // 発散
  ll = a * sdlh + (double)nd * la + (a - 1.0) * sdlt - shta;
  gv.assign(0, (double)nd + a * (sdlh + sdlt - shtalh - shtalt));
  hmat.assign(0, 0, a * (sdlh + sdlt - shtalh - a * shtalh2 - 2.0 * a * shtalhlt - shtalt - a * shtalt2));
  for (int j = 0; j < nxt; ++j)
  { gv.add(j + 1, a * sdxv(j));  hmat.add(0, j + 1, a * sdxv(j)); }
  for (int i = 0; i < npt - 1; ++i)
  {
    for (int j = i + 1; j < npt; ++j) hmat.assign(j, i, hmat(i, j));
  }
  return reslgh;                // 計算成功
}

void varclass::cal_pltab(option& opt) // 分散分析表用各種統計量を計算する
{
  vdft = 0;  vdfb = vcheckvec[vnp] - 1;
  vl0 = 0.0;  vlf = 0.0;
  if (opt.model() == modelwei)  // aと切片項の対数尤度を求める
  {
    vdfb -= 1;
    vecdbl tbvec(vnp, 0.0), tvbvec(vnp, 0.0);
    vecint tcheckvec(vnp + 1, 0);
    tbvec[0] = v_bvec[0];  tbvec[1] = v_bvec[1];
    tcheckvec[0] = tcheckvec[1] = 1; // aと切片項のみ選択
    tcheckvec[vnp] = 2;
    int nr = 0;
    newton_pltab(opt, tbvec, tvbvec, nr, vl0, tcheckvec);
  }
  else
  {
    vl0 = (double)vnd * (log(vlamda0) - 1.0);
    vdft = cal_pltablf(opt, vlf, vcheckvec);
  }
  vdfl = vdft - vdfb;
  vlt = vlf - vl0;  vlreg = vll - vl0;  vllof = vlf - vll;
  if (vdft < 1 || vlt < FDeps) vlt = 0.0;
  if (vdfb < 1 || vlreg < FDeps) vlreg = 0.0;
  if (vdfl < 1 || vllof < FDeps) vllof = 0.0;
}

int varclass::cal_pltablf(option& opt, double& lf, const vecint& checkvec) const
// 飽和モデルの尤度を求める
{
  int is = (opt.model() == modelwei) ? 2 : 1;
  int nxt = checkvec[vnp] - is;
  vector<bool> lcvec(vlifemlst.size(), true);
  vec<double> xiv(nxt), xjv(nxt);
  int nlf = 0, lci = 0;
  lf = 0.0;
  for (listlfm_citr iitr = vlifemlst.begin(); iitr != vlifemlst.end(); ++iitr, ++lci)
  {
    if (lcvec[lci] == true)
    {
      int k = 0, j = 0;
      for (int i = is; i < vnp; ++i, ++j)
      {
        if (checkvec[i])
        {
          xiv.assign(k, iitr->x(j));
          ++k;
        }
      }
      double sumti = iitr->t();
      int ni = 1, ndi = 0;
      if (iitr->oc() == ocdeath) ndi = 1; // 死亡
      listlfm_citr jitr = iitr;
      ++jitr;
      int lcj = lci + 1;
      while (jitr != vlifemlst.end())
      {
        if (lcvec[lcj] == true)
        {
          k = 0;  j = 0;
          for (int i = is; i < vnp; ++i, ++j)
          {
            if (checkvec[i])
            {
              xjv.assign(k, jitr->x(j));
              ++k;
            }
          }
          if (xiv == xjv)       // xが等しい
          {
            lcvec[lcj] = false; // 1度しか数えない
            sumti += jitr->t();
            ++ni;
            if (jitr->oc() == ocdeath) ++ndi;
          }
        }
        ++jitr;  ++lcj;
      }
      if (ndi > 0)
      {
        nlf += ni - 1;
        if (sumti > 0.0)
        {
          double hi = (double)ndi / sumti;
          lf += (double)ndi * (log(hi) - 1.0);
        }
      }
    }
  }
  return vnc - 1 - nlf;         // 全体の自由度を返す
}

class pltabclass                // pltabメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 分析結果のタイトルとデータ識別情報を出力する
  void put_ltab(ostream& fout, option& opt, const varclass& var) const;
  // Kaplan-Meier法による生命表を出力する
  void put_mst(ostream& fout, option& opt, double sp, double sn, double tp, double tn, double tlow, double tup, double tmax) const;
  // %生存期間推定値を出力する
  void put_xmsd(ostream& fout, option& opt, const idd& id, const varclass& var) const;
  // 共変数の基礎統計量を出力する
  void put_pltab(ostream& fout, option& opt, ldata& datalst, bool stepsel, const varclass& var) const;
  // パラメトリックモデルの解析結果を出力する
  void put_b(ostream& fout, option& opt, const varclass& var) const;
  // 偏回帰係数を出力する
  void put_bci(ostream& fout, option& opt, const varclass& var) const;
  // 偏回帰係数の信頼区間を出力する
  void put_plttab(ostream& fout, option& opt, ldata& datalst, const varclass& var) const;
  // パラメトリックモデルの分散分析表を出力する
  void put_cindex(ostream& fout, option& opt, ldata& datalst, const varclass& var) const;
  // C-indexを出力する
  int cal_cindex(option& opt, int& nci, double& u, double& ci, ldata& hrlst) const;
  // C-indexを計算する
  void put_dns(option& opt, idd& id, ldata& datalst, bool stepsel, const varclass& var) const;
  // 理論ハザードと生存率をデータファイル形式で出力する
  void put_dnsid(ofstream& fout, option& opt, idd& id, bool stepsel) const;
  // DANSデータのタイトルとデータ識別情報を出力する
  void plot_ngp(option& opt, idd& id, ldata& datalst, bool stepsel, const varclass& var) const;
  // Ngraph用出力
  void plot_ngpid(ofstream& fout, option& opt, idd& id, bool stepsel, const varclass& var) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_hr(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用にHRとその信頼区間を出力する
  void plot_kms(ofstream& fout, const varclass& var) const;
  // Ngraph用にKaplan-Meier法による累積生存率曲線を出力する
  void plot_kmd(ofstream& fout, const varclass& var) const;
  // Ngraph用にKaplan-Meier法による累積死亡率曲線を出力する
  void plot_sur(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に生存率関数を出力する
  void plot_dea(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に死亡率関数を出力する
public:
  pltabclass() : fname()        // コンストラクタ
  {}
  ~pltabclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void pltabclass::put_resid(ostream& fout, idd& id) const
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "期間項目  :" << id.item(0) << "\n";
  fout << "転帰項目  :" << id.item(1) << "\n";
  for (size_t i = 2; i < id.ni(); ++i)
  {
    fout << "共変数x" << setw(3) << i - 1;
    fout << ":" << id.item(i) << "\n";
  }
}

void pltabclass::put_ltab(ostream& fout, option& opt, const varclass& var) const
// Kaplan-Meier法による生命表を出力する
{
  constexpr int LENline = 75;   // 区切り線の長さ
  fout << "\n・Kaplan-Meier法による累積生存率    死亡コード:" << opt.dcode() << "  +:打ち切り (+):脱落\n";
  fout << "番号  ケース番号  生存期間(転帰)  生存例数/観測例数  累積生存率    標準誤差\n";
  putline(fout, LENline);
  double sigw = 0.0, se = 0.0, sp = 1.0, sn = 1.0, zt = 0.0;
  double tmax = 0.0, tp = 0.0, tn = 0.0, tlow = 0.0, tup = 0.0;
  int nc = var.nc(), i = 0;
  double z = ptoz(1.0 - opt.cc());
  for (auto& lifemi : var.lifemlst())
  {
    fout << setw(4) << i + 1;
    fout << setw(8) << lifemi.cno();
    fout << setw(14) << lifemi.t();
    if (lifemi.oc() == ocdeath)
    {
      if (i + 1 < nc) sigw += 1.0 / (double)(nc - i) / (double)(nc - i - 1);
      se = lifemi.s() * sqrt(sigw);
      fout << setw(16) << nc - i - 1;
      fout << "/" << left << setw(8) << nc - i << right;
      fout << setw(12) << lifemi.s() << setw(12) << se << "\n";
      zt = (se > 0.0) ? (lifemi.s() - opt.ps()) / se : 0.0;
      if (zt >= z) tlow = lifemi.t();
      if (tup == 0.0 && -z >= zt) tup = lifemi.t();
    }
    else
    {
      fout << left;
      if (lifemi.oc() == ocalive) fout << setw(10) << " +";
      else fout << setw(10) << "(+)";
      fout << right << "(" << setw(5) << nc - i;
      fout << "/" << left << setw(5) << nc - i << right;
      fout << ")" << setw(14) << lifemi.s() << setw(12) << se << "\n";
    }
    if (lifemi.s() >= opt.ps()) { tp = lifemi.t();  sp = lifemi.s(); }
    if (tn == 0.0 && opt.ps() >= lifemi.s())
    { tn = lifemi.t();  sn = lifemi.s(); }
    tmax = lifemi.t();
    ++i;
  }
  putline(fout, LENline);
  if (opt.res("ps")) put_mst(fout, opt, sp, sn, tp, tn, tlow, tup, tmax);
}

void pltabclass::put_mst(ostream& fout, option& opt, double sp, double sn, double tp, double tn, double tlow, double tup, double tmax) const
// %生存期間推定値を出力する
{
  fout << "  " << opt.ps() * 100.0 << "%生存時間推定値=";
  if (tn == 0.0)
  { fout << tmax << "以上"; }
  else
  {
    if (sp > sn) tp += (tn - tp) * (sp - opt.ps()) / (sp - sn);
    fout << tp;
  }
  fout << "    " << opt.cc() * 100.0 << "%信頼区間=" << tlow << "〜";
  if (tup == 0.0) fout << tmax << "以上";
  else fout << tup;
  fout << endl;
}

void pltabclass::put_xmsd(ostream& fout, option& opt, const idd& id, const varclass& var) const
// 共変数の基礎統計量を出力する
{
  fout << "\n・共変数の基礎統計量:y=ln(" << id.item(0) << ")\n";
  constexpr int LENline = 72;
  putline(fout, LENline);
  size_t yidx = var.nx();
  fout << "y   :";
  putmsd(fout, var.n(yidx), var.mean(yidx), var.vr(yidx));
  for (size_t i = 1; i < yidx; ++i)
  {
    fout << "x" << setw(3) << i;
    fout << ":";
    putmsd(fout, var.n(i), var.mean(i), var.vr(i));
  }
  putline(fout, LENline);
  fout << "死亡コード=" << opt.dcode() << "  死亡数=" << var.nd() << "  打ち切り数=" << var.nc() - var.nd() << "  平均瞬間死亡率=" << var.lamda0() << endl;
}

void pltabclass::put_pltab(ostream& fout, option& opt, ldata& datalst, bool stepsel, const varclass& var) const
// パラメトリックモデルの解析結果を出力する
{
  if (stepsel) fout << "\n・前進的変数増減法による変数選択結果";
  else fout << "\n・全変数を選択した結果";
  if (opt.method() == methoddel) fout << "  反復回数(差分法)=";
  else fout << "  反復回数(微分法)=";
  fout << var.nr();
  if (var.resvar() == reserr) fout << ":Newton法発散！";
  else if (var.resvar() == resrmax) fout << "(反復回数最大値)";
  fout << "\n";
  if (stepsel)
  {
    if (opt.pin() < 1.0)
    { fout << "  取り込み基準:p値≦" << opt.pin() << "  追い出し基準:p値＞" << opt.pin(); }
    else
    { fout << "  取り込み基準:χ2乗≧" << opt.cin() << "  追い出し基準:χ2乗＜" << opt.cin(); }
    fout << "\n";
  }
  int is = 0;
  double la = 0.0, a = 1.0;
  if (opt.model() == modelwei)
  {
    fout << "  Weibull分布モデル:S(t)=exp{-(λt)^a}=exp{-exp(β0+Σβjxj)(t^a)}\n";
    is = 1;  la = var.b(0);  a = exp(var.b(0));
  }
  else
  { fout << "  指数分布モデル:S(t)=exp(-λt)=exp{-exp(β0+Σβjxj)t}\n"; }
  fout << "  S(t):生存関数  β0:切片  βj:共変数xjの偏回帰係数\n";
  if (opt.model() == modelwei && var.resvar() != reserr)
  { fout << "  a=" << a << "  Scale=1/a=" << 1.0 / a << "\n"; }
  if (opt.res("v"))
  {
    fout << "  初期値:";
    if (opt.model() == modelwei) fout << "ln(a)=" << var.b0(0) << " ";
    for (int i = is; i < var.np(); ++i) fout << "β" << i - is << "=" << var.b0(i) << " ";
    fout << "\n";
  }
  if (var.resvar() == reserr) return;
  put_b(fout, opt, var);
  put_bci(fout, opt, var);
  put_plttab(fout, opt, datalst, var);
}

void pltabclass::put_b(ostream& fout, option& opt, const varclass& var) const
// 偏回帰係数を出力する
{
  fout << "                                            標準                   有意確率\n";
  fout << "共変数 偏回帰係数  標準誤差  ハザード比  偏回帰係数  Waldのχ2乗       p値\n";
  int llen = 78;
  putline(fout, llen);
  int is = 0;
  double csqr = 0.0, bi = 0.0, sei = 0.0;
  ldbl pr = 0.0;
  if (opt.model() == modelwei)
  {
    csqr = (var.vb(0) > 0.0) ? var.b(0) * var.b(0) / var.vb(0) : 0.0;
    pr = ctop(csqr, 1);  bi = var.b(0);  sei = sqrt(var.vb(0));
    fout << "ln(a)" << setw(11) << bi << setw(11) << sei;
    fout << setw(36) << csqr << setw(12) << pr << pind(pr) << "\n";
    is = 1;
  }
  csqr = (var.vb(is) > 0.0) ? var.b(is) * var.b(is) / var.vb(is) : 0.0;
  pr = ctop(csqr, 1);  bi = var.b(is);  sei = sqrt(var.vb(is));
  fout << "切片" << setw(12) << bi << setw(11) << sei;
  fout << setw(36) << csqr << setw(12) << pr << pind(pr) << "\n";
  for (int i = is + 1; i < var.np(); ++i)
  {
    if (var.check(i))
    {
      bi = var.b(i);  sei = sqrt(var.vb(i));
      csqr = (var.vb(i) > 0.0) ? var.b(i) * var.b(i) / var.vb(i) : 0.0;
      pr = ctop(csqr, 1);
      fout << "x" << setw(3) << i - is << setw(12) << bi;
      fout << setw(11) << sei << setw(12) << exp(bi);
      fout << setw(12) << bi * var.sd(i);
      fout << setw(12) << csqr << setw(12) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, llen);
}

void pltabclass::put_bci(ostream& fout, option& opt, const varclass& var) const
// 偏回帰係数の信頼区間を出力する
{
  fout << "\n共変数  偏回帰係数" << setw(5) << opt.cc() * 100.0;
  fout << "%CI下限        上限  ハザード比" << setw(5) << opt.cc() * 100.0;
  fout << "%CI下限        上限\n";
  constexpr int Llen = 78;
  putline(fout, Llen);
  int is = 0;
  double la = 0.0, a = 1.0;
  double bi = 0.0, sei = 0.0, z = 0.0, ci = 0.0, bl = 0.0, bu = 0.0;
  if (opt.model() == modelwei)
  {
    is = 1;  la = var.b(0);  a = exp(var.b(0));
    bi = var.b(0);  sei = sqrt(var.vb(0));  z = ptoz(1.0 - opt.cc());
    ci = z * sei;
    bl = bi - ci; bu = bi + ci;
    fout << "ln(a)" << setw(13) << bi << setw(12) << bl;
    fout << setw(12) << bu << "\n";
  }
  bi = var.b(is);  sei = sqrt(var.vb(is));
  z = ptoz(1.0 - opt.cc());
  ci = z * sei;
  bl = bi - ci;  bu = bi + ci;
  fout << "切片" << setw(14) << bi << setw(12) << bl << setw(12) << bu << "\n";
  for (int i = is + 1; i < var.np(); ++i)
  {
    if (var.check(i))
    {
      bi = var.b(i);  sei = sqrt(var.vb(i));
      ci = z * sei;
      bl = bi - ci;  bu = bi + ci;
      fout << "x" << setw(3) << i - is << setw(14) << bi;
      fout << setw(12) << bl << setw(12) << bu << setw(12) << exp(bi);
      fout << setw(12) << exp(bl) << setw(12) << exp(bu) << "\n";
    }
  }
  putline(fout, Llen);
}

void pltabclass::put_plttab(ostream& fout, option& opt, ldata& datalst, const varclass& var) const
// パラメトリックモデルの分散分析表を出力する
{
  double r2 = (var.lt() > 0.0) ? var.lreg() / var.lt() : 0.0;
  double csqr = 2.0 * var.lreg();
  ldbl pr = ctop(csqr, var.dfb());
  fout << "\n対数尤度:L(モデル)=" << var.ll() << "  L(切片)=" << var.l0();
  if (opt.model() == modelwei)
  {
    fout << "  AIC=" << 2.0 * ((double)(var.dfb() + 1) - var.ll()) << "\n";
    fout << "回帰の尤度比検定:χ2乗=" << csqr << "  自由度=" << var.dfb() << "  有意確率p=" << pr << pind(pr) << "\n";
    return;
  }
  fout << "  飽和Lf=" << var.lf() << "\n";
  fout << "擬似寄与率:R2乗=" << r2 << "  AIC=" << 2.0 * ((double)(var.dfb() + 1) - var.ll()) << "\n";
  if (opt.res("cindex")) put_cindex(fout, opt, datalst, var);
  const string Head = "回帰とズレの検定";
  int w = strclen(Head, CHARwide);
  int npad = Head.size() - w;
  fout << "\n" << setw((68 + w) / 2 + npad) << Head;
  fout << "\n要因         -(対数尤度比)    自由度          χ2乗   有意確率p値\n";
  constexpr int LENline = 68;
  putline(fout, LENline);
  fout << "回帰" << setw(22) << var.lreg() << setw(10) << var.dfb();
  fout << setw(14) << csqr << setw(14) << pr << pind(pr) << "\n";
  csqr = 2.0 * var.llof();
  pr = ctop(csqr, var.dfl());
  fout << "ズレ(LOF)" << setw(17) << var.llof() << setw(10) << var.dfl();
  fout << setw(14) << csqr << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  fout << "全体" << setw(22) << var.lt() << setw(10) << var.dft() << endl;
}

void pltabclass::put_cindex(ostream& fout, option& opt, ldata& datalst, const varclass& var) const
// C-indexを出力する
{
  ldata hrlst;                  // C-index計算用データリスト
  datac cdata(3);
  int is = (opt.model() == modelwei) ? 1 : 0;
  vecdbl xvec(var.nx() - 1);
  for (auto& cdatai : datalst)
  {
    double lh = BLANKf, t = cdatai.fdata(0);
    bool blank = (t == BLANKf) ? true : false;
    for (int j = 1; blank == false && j < var.nx(); ++j)
    {
      xvec[j - 1] = cdatai.fdata(j + 1);
      if (xvec[j - 1] == BLANKf) blank = true;
    }
    if (blank == false)
    {
      lh = var.b(is);
      int j = 0;
      for (int i = is + 1; i < var.np(); ++i, ++j)
      {
        if (var.check(i)) lh += var.b(i) * xvec[j];
      }
    }
    cdata.assign(0, cdatai.sdata(0));  cdata.assign(1, cdatai.sdata(1));
    cdata.assign(2, lh);
    hrlst.push_back(cdata);
  }
  int nci = 0;
  double u = 0.0, ci = 0.0;
  int n = cal_cindex(opt, nci, u, ci, hrlst);
  fout << "C-index=" << u << "/" << nci << "=" << ci << "  1-C-index=" << 1.0 - ci << endl;
}

int pltabclass::cal_cindex(option& opt, int& nci, double& u, double& ci, ldata& hrlst) const
// C-indexを計算する
{
  nci = 0;  u = 0.0;  ci = 0.0;
  int n = 0;
  for (ldata_itr iitr = hrlst.begin(); iitr != hrlst.end(); ++iitr)
  {
    double ti = iitr->fdata(0), fi = iitr->fdata(2);
    ++n;
    octype oi = ocalive;                               // 生存
    if (iitr->sdata(1) == opt.dcode()) oi = ocdeath;   // 死亡
    ldata_itr jitr = iitr;
    ++jitr;
    while (jitr != hrlst.end())
    {
      if (iitr != jitr)
      {
        double tj = jitr->fdata(0), fj = jitr->fdata(2);
        if (tj != BLANKf && fj != BLANKf)
        {
          octype oj = ocalive;                             // 生存
          if (jitr->sdata(1) == opt.dcode()) oj = ocdeath; // 死亡
          if (oi == ocdeath && oj == ocdeath)
          {
            if (ti < tj)
            {
              ++nci;
              if (fi > fj) u = u + 1.0;
              else if (fi == fj) u = u + 0.5;
            }
            else if (ti > tj)
            {
              ++nci;
              if (fi < fj) u = u + 1.0;
              else if (fi == fj) u = u + 0.5;
            }
          }
          else if (oi == ocdeath && oj == ocalive)
          {
            if (ti <= tj)
            {
              ++nci;
              if (fi > fj) u = u + 1.0;
              else if (fi == fj) u = u + 0.5;
            }
          }
          else if (oi == ocalive && oj == ocdeath)
          {
            if (ti >= tj)
            {
              ++nci;
              if (fi < fj) u = u + 1.0;
              else if (fi == fj) u = u + 0.5;
            }
          }
        }
      }
      ++jitr;
    }
  }
  if (nci > 0) ci = u / (double)nci;
  return n;                     // 評価例数を返す
}

void pltabclass::put_dns(option& opt, idd& id, ldata& datalst, bool stepsel, const varclass& var) const
// 対数ハザードと生存率をデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout, opt, id, stepsel);
  datac cdata(4);
  int is = 0;
  double la = 0.0, a = 1.0;
  if (opt.model() == modelwei)
  {
    is = 1;  la = var.b(0);
    a = exp(la);
  }
  vecdbl xvec(var.nx() - 1);
  for (auto& cdatai : datalst)
  {
    cdata.assign(0, cdatai.sdata(0));  cdata.assign(1, cdatai.sdata(1));
    double lh = BLANKf, h = BLANKf, sc = BLANKf, t = cdatai.fdata(0);
    bool blank = (t == BLANKf) ? true : false;
    for (int j = 1; blank == false && j < var.nx(); ++j)
    {
      xvec[j - 1] = cdatai.fdata(j + 1);
      if (xvec[j - 1] == BLANKf) blank = true;
    }
    if (blank == false)
    {
      lh = var.b(is);
      int j = 0;
      for (int i = is + 1; i < var.np(); ++i, ++j)
      {
        if (var.check(i)) lh += var.b(i) * xvec[j];
      }
      h = exp(lh);
      sc = exp(-pow(h * t, a));
    }
    cdata.assign(2, lh);  cdata.assign(3, sc);
    cdata.put(fout);
  }
}

void pltabclass::put_dnsid(ofstream& fout, option& opt, idd& id, const bool stepsel) const
// DANSデータのタイトルとデータ識別情報を出力する
{
  idd tid(id.title(), 4);
  tid.assign_item(0, id.item(0));
  tid.assign_item(1, id.item(1));
  if (opt.model() == modelwei)
  {
    if (stepsel)
    {
      tid.assign_item(2, "Weibull分布モデルの対数ハザード(変数選択法)");
      tid.assign_item(3, "Weibull分布モデルの理論生存率(変数選択法)");
    }
    else
    {
      tid.assign_item(2, "Weibull分布モデルの対数ハザード(全変数選択)");
      tid.assign_item(3, "Weibull分布モデルの理論生存率(全変数選択)");
    }
  }
  else
  {
    if (stepsel)
    {
      tid.assign_item(2, "指数分布モデルの対数ハザード(変数選択法)");
      tid.assign_item(3, "指数分布モデルの理論生存率(変数選択法)");
    }
    else
    {
      tid.assign_item(2, "指数分布モデルの対数ハザード(全変数選択)");
      tid.assign_item(3, "指数分布モデルの理論生存率(全変数選択)");
    }
  }
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  tid.put_item(fout, formtag);
}

void pltabclass::plot_ngp(option& opt, idd& id, ldata& datalst, bool stepsel, const varclass& var) const
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("sur"))
  {
    fout << "-s1 -dline -w80 -zy0,1,0.2\n";
    fout << "%脱落例を描く時は-derrorbar_y -w40 -o0 -zy0,1,0.2\n";
    plot_ngpid(fout, opt, id, stepsel, var);
    plot_sur(fout, opt, var);
    delml = true;
  }
  if (opt.ngp("dea"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      fout << "%脱落例を描く時は-derrorbar_y -w40 -o0 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, opt, id, stepsel, var);
    plot_dea(fout, opt, var);
  }
  if (opt.ngp("kms"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      fout << "%脱落例を描く時は-derrorbar_y -w40 -o0 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, opt, id, stepsel, var);
    plot_kms(fout, var);
  }
  if (opt.ngp("kmd"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      fout << "%脱落例を描く時は-derrorbar_y -w40 -o0 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, opt, id, stepsel, var);
    plot_kmd(fout, var);
  }
  if (opt.ngp("hr"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -w80 -o400 -zx0,"<< id.ni() - 1 << ",1 -zy0.1,100,10 -eylog\n";
      fout << "%CIを描く時は-s1 -derrorbar_y -w80\n";
    }
    plot_ngpid(fout, opt, id, stepsel, var);
    plot_hr(fout, opt, var);
  }
}

void pltabclass::plot_ngpid(ofstream& fout, option& opt, idd& id, bool stepsel, const varclass& var) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%期間項目  :" << id.item(0) << "\n";
  fout << "%転帰項目  :" << id.item(1) << "\n";
  int is = 0, k = 0;
  if (opt.model() == modelwei) is = 1;
  for (int i = is + 1; i < var.np(); ++i, ++k)
  {
    if (var.check(i))
    { fout << "%共変数x" << setw(3) << k + 1 << ":" << id.item(k + 2) << "\n"; }
  }
  if (stepsel)
  {
    if (opt.pin() < 1.0) fout << "%取り込み基準:p値≦" << opt.pin() << "\n";
    else fout << "%取り込み基準:Waldのχ2乗≧" << opt.cin() << "\n";
  }
  else
  { fout << "%全変数選択\n"; }
  k = 0;
  for (int i = is + 1; i < var.np(); ++i, ++k)
  {
    if (var.check(i))
    {
      if (opt.x(k) != BLANKf) fout << "%値指定共変数:x" << k + 1 << "=" << opt.x(k) << "\n";
    }
  }
}

void pltabclass::plot_hr(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用にHRとその信頼区間を出力する
{
  int is = 0;
  if (opt.model() == modelwei)
  {
    is = 1;
    fout << "%Weibull分布モデルのハザード比と信頼区間\n";
  }
  else
  { fout << "%指数分布モデルのハザード比と信頼区間\n"; }
  double z = ptoz(1.0 - opt.cc());
  for (int i = is + 1; i < var.np(); ++i)
  {
    if (var.check(i))
    {
      double ci = z * sqrt(var.vb(i));
      double lhri = var.b(i);
      double hri = exp(lhri);
      double hrl = exp(lhri - ci), hru = exp(lhri + ci);
      fout << i - is << " " << hri << " " << hru - hri << " " << hrl - hri << "\n";
    }
  }
}

void pltabclass::plot_kms(ofstream& fout, const varclass& var) const
// Ngraph用にKaplan-Meier法による累積生存率曲線を出力する
{
  fout << "%Kaplan-Meier法による累積生存率曲線\n";
  fout << "0 1\n";
  double spre = 1.0;
  for (auto& lifemi : var.lifemlst())
  {
    fout << lifemi.t() << " " << spre;
    if (lifemi.oc() == ocdeath)
    { fout << "\n" << lifemi.t() << " " << lifemi.s() << "\n"; }
    else
    { fout << " 0 0.02\n"; }
    spre = lifemi.s();
  }
}

void pltabclass::plot_kmd(ofstream& fout, const varclass& var) const
// Ngraph用にKaplan-Meier法による累積死亡率曲線を出力する
{
  fout << "%Kaplan-Meier法による累積死亡率曲線\n";
  fout << "0 0\n";
  double spre = 1.0;
  for (auto& lifemi: var.lifemlst())
  {
    fout << lifemi.t() << " " << 1.0 - spre;
    if (lifemi.oc() == ocdeath)
    { fout << "\n" << lifemi.t() << " " << 1.0 - lifemi.s() << "\n"; }
    else
    { fout << " 0 0.02\n"; }
    spre = lifemi.s();
  }
}

void pltabclass::plot_sur(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に生存率関数を出力する
{
  int is = 0, k = 0;
  double la = 0.0, a = 1.0;
  if (opt.model() == modelwei)
  {
    is = 1;  la = var.b(0);
    a = exp(la);
    fout << "%Weibull分布モデルの生存率関数\n";
  }
  else
  { fout << "%指数分布モデルの生存率関数\n"; }
  double lh = var.b(is);
  for (int i = is + 1; i < var.np(); ++i, ++k)
  {
    if (var.check(i))
    {
      double xk = (opt.x(k) == BLANKf) ? var.mean(k + 1) : opt.x(k);
      lh += var.b(i) * xk;
    }
  }
  double h = exp(lh);
  constexpr uint RDdef = 50;    // X軸の分割数
  double tmin = 0.0, tmg = 0.1;
  listlfm_citr litr = (var.lifemlst()).end();
  --litr;                       // 生命表の最後の要素
  double tmax = litr->t();      // 時期最大値
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf)
  { tmin = opt.xmin();  tmax = opt.xmax();  tmg = 0.0; }
  double ltmax = log(tmax);
  double ltmin = (tmin > 0) ? log(tmin) : ltmax / (double)(10 * RDdef);
  double ltend = ltmax + (ltmax - ltmin) * tmg;
  double ldt = (ltend - ltmin) / (double)RDdef;
  double lt = ltmin;
  if (tmin == 0.0) fout << "0 1\n";
  for (uint i = 0; i <= RDdef; ++i, lt += ldt)
  {
    double t = exp(lt);
    fout << t << " " << exp(-pow(h * t, a)) << "\n";
  }
}

void pltabclass::plot_dea(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に死亡率関数を出力する
{
  int is = 0, j = 0;
  double a = 1.0;
  if (opt.model() == modelwei)
  {
    is = 1;  a = var.b(0);
    fout << "%Weibull分布モデルの死亡率関数\n";
  }
  else
  { fout << "%指数分布モデルの死亡率関数\n"; }
  double lh = var.b(is);
  for (int i = is + 1; i < var.np(); ++i, ++j)
  {
    if (var.check(i))
    {
      double xj = (opt.x(j) == BLANKf) ? var.mean(j + 1) : opt.x(j);
      lh += var.b(i) * xj;
    }
  }
  double h = exp(lh);
  constexpr uint RDdef = 50;    // X軸の分割数
  double tmin = 0.0, tmg = 0.1;
  listlfm_citr litr = (var.lifemlst()).end();
  --litr;                       // 生命表の最後の要素
  double tmax = litr->t();      // 時期最大値
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf)
  { tmin = opt.xmin();  tmax = opt.xmax();  tmg = 0.0; }
  double ltmax = log(tmax);
  double ltmin = (tmin > 0) ? log(tmin) : ltmax / (double)(10 * RDdef);
  double ltend = ltmax + (ltmax - ltmin) * tmg;
  double ldt = (ltend - ltmin) / (double)RDdef;
  double lt = ltmin;
  if (tmin == 0.0) fout << "0 0\n";
  for (uint i = 0; i <= RDdef; ++i, lt += ldt)
  {
    double t = exp(lt);
    fout << t << " " << 1.0 - exp(-pow(h * t, a)) << "\n";
  }
}

int pltabclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw option_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void pltabclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  varclass var(opt, id.ni());            // 変数オブジェクト
  var.cal(opt, datalst);                 // 計算を実行する
  if (opt.res("res"))
  {
    if (opt.res("ltab"))
    {
      put_resid(cout, id);
      if (var.resvar() == reserr)
      {
        cout << "※生命表データなし！\n\n" << endl;
        return;
      }
      put_ltab(cout, opt, var);
      cout << "\n\n";
    }
    put_resid(cout, id);
    if (opt.res("stat"))
    {
      put_xmsd(cout, opt, id, var);
      putmat(cout, "・相関行列:y1=ln("+id.item(0)+")", matr, var.nc(), var.smat(), var.nx() - 1, 1);
    }
    if (var.nc() < 1 + var.nx()) // 重回帰モデル計算できず
    {
      cout << "※重回帰モデル計算不可能！\n\n" << endl;
      return;
    }
  }
  bool putfile = false;
  if (opt.cin() > 0.0 || (0.0 < opt.pin() && opt.pin() < 1.0)) // 変数選択
  {
    var.stepsel(opt);
    if (opt.res("selx")) put_pltab(cout, opt, datalst, true, var);
    if (var.resvar() != reserr)
    {
      if (opt.dnsname() != "") put_dns(opt, id, datalst, true, var);
      if (opt.ngpname() != "") plot_ngp(opt, id, datalst, true, var);
      putfile = true;
    }
    if (var.nb() >= var.np() && opt.res("allx") == false) // 全変数を選択した時
    {
      if (opt.res("res")) cout << "\n" << endl;
      return;
    }
  }
  var.allsel(opt);
  if (opt.res("def") || opt.res("allx")) put_pltab(cout, opt, datalst, false, var);
  if (opt.res("res")) cout << "\n" << endl;
  if (var.resvar() != reserr && putfile == false)
  {
    if (opt.dnsname() != "") put_dns(opt, id, datalst, false, var);
    if (opt.ngpname() != "") plot_ngp(opt, id, datalst, false, var);
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  pltabclass pltab;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) pltab.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { pltab.get(argv[i], opt, id, datalst);  fok = false; }
    }
    pltab.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
